#!/usr/bin/env python
import lz4.frame
import dill
import os

def main():
    cwd = os.getcwd()
    with lz4.frame.open("task.p.lz4", 'rb') as f:
        task = dill.load(f)

    print("Task = {}\n\nargs = {}\n\nkwargs = {}\n".format(
        task["task"], task["args"], task["kwargs"],
    ))
    result = task["task"](*task["args"], **task["kwargs"])

    # Just incase the user wants to change directory within the task
    os.chdir(cwd)

    with lz4.frame.open("result.p.lz4", 'wb') as f:
        dill.dump(result, f)

if __name__ == "__main__":
    main()



import os
import re
import logging
import time
from builtins import input
from multiprocessing import Pool
from tqdm.auto import tqdm

from .utils import run_command

logger = logging.getLogger(__name__)

class SGETaskSubmitter(object):
    submit_command = 'qsub -wd {wd} -V -e /dev/null -o /dev/null -t {start}-{njobs}:1 {job_opts} {executable}'
    regex_submit = re.compile('Your job-array (?P<jobid>[0-9]+)\.(?P<start>[0-9]+)-(?P<stop>[0-9]+):1 \(".*"\) has been submitted')
    def __init__(self, job_options):
        self.job_options = job_options
        self.jobid_tasks = {}

    def submit_tasks(
        self, tasks, start=0, dryrun=False, request_user_input=False,
        quiet=False,
    ):
        if tasks is None or len(tasks) <= 0:
            return

        curdir = os.getcwd()
        njobs = len(tasks)
        executable = run_command("which pysge_worker.sh")[0].decode("utf-8")

        job_opts = self.job_options
        if request_user_input:
            job_opts = input(
                "Using job options '{}'. Insert new options or nothing to use "
                "the default\n:".format(job_opts)
            )
            job_opts = job_opts if job_opts != "" else self.job_options

        cmd = self.submit_command.format(
            executable=executable, start=start+1, njobs=njobs+start,
            job_opts=job_opts, wd=os.path.dirname(tasks[0]),
        )
        if not dryrun:
            out, err = run_command(cmd)
            match = self.regex_submit.search(out.decode("utf-8"))
            if match is None:
                logger.error("Command: {}".format(cmd))
                logger.error("Malformed qsub submission string: {}".format(repr(out.decode("utf-8"))))
                logger.error("Skipping and returning on the next resubmission round")
                return
            jobid = int(match.group("jobid"))
            start = int(match.group("start"))
            stop = int(match.group("stop"))
            logger.info('Submitted {}.{}-{}:1'.format(jobid, start, stop))
        else:
            print(cmd)
            jobid = 0

        for aid in range(njobs):
            self.jobid_tasks['{}.{}'.format(jobid, aid+start)] = tasks[aid]

    def killall(self):
        jids = []
        for jobid, _ in self.jobid_tasks.items():
            tjid = jobid.split(".")[0]
            if tjid not in jids:
                jids.append(tjid)
        cmd = "qdel {}".format(" ".join(jids))
        run_command(cmd)

class MPTaskSubmitter(object):
    def submit_tasks(self, tasks, ncores=4, sleep=1, quiet=False):
        if tasks is None or len(tasks) <= 0:
            return

        results = []
        pool = Pool(processes=ncores)
        pbar = tqdm(total=len(tasks), desc="Finished", disable=quiet)

        try:
            for task in tasks:
                results.append(
                    pool.apply_async(task['task'], task['args'], task['kwargs'])
                )

            nremaining = len(tasks)
            while nremaining > 0:
                nfinished = sum(r.ready() for r in results)
                nremaining = len(tasks) - nfinished
                pbar.n = nfinished
                pbar.refresh()
                time.sleep(sleep)

            results = [r.get() for r in results]
        except KeyboardInterrupt:
            pool.terminate()
            results = []
        pool.close()
        pbar.close()
        print("")

        return results



import os
import dill
import lz4.frame
import logging
from tqdm.auto import tqdm
from .area import WorkingArea
from .submitter import SGETaskSubmitter, MPTaskSubmitter
from .monitor import JobMonitor

logger = logging.getLogger(__name__)

def _validate_tasks(tasks):
    for task in tasks:
        if any(key not in task for key in ("task", "args", "kwargs")):
            return False
        if not callable(task["task"]):
            return False
    return True

def sge_submit(
    tasks, label, tmpdir, options="-q hep.q", dryrun=False, quiet=False,
    sleep=5, request_resubmission_options=True, return_files=False,
    dill_kw={"recurse": True},
):
    """
    Submit jobs to an SGE batch system. Return a list of the results of each
    job (i.e. the return values of the function calls)

    Parameters
    ----------
    tasks : list
        A list of dictrionaries with the keys: task, args and kwargs. Each
        element is run on a node as task(*args, **kwargs).

    label : str
        Label given to the qsub submission script through -N.

    tmpdir : str
        Path to temporary directory (doesn't have to exist) where pysge stores
        job infomation. Each call will have a unique identifier in the form
        tpd_YYYYMMDD_hhmmss_xxxxxxxx. Within this directory exists all tasks in
        separate directories with a dilled file, stdout and stderr for that
        particular job.

    options : str (default = "-q hep.q")
        Additional options to pass to the qsub command. Take care since the
        following options are already in use: -wd, -V, -e, -o and -t.

    dryrun : bool (default = False)
        Create directories and files but don't submit the jobs.

    quiet : bool (default = False)
        Don't print tqdm progress bars. Other prints are controlled by logging.

    sleep : float (default = 5)
        Minimum time between queries to the batch system.

    request_resubmission_options : bool (default = True)
        When a job fails the master process will expect an stdin from the user
        to alter the submission options (e.g. to increase walltime or memory
        requested). If False it will use the original options.

    return_files : bool (default = False)
        Instead of opening the output files and loading them into python, just
        send the paths to the output files and let the user deal with them.

    dill_kw : dict
        Kwargs to pass to dill.dump
    """
    if not _validate_tasks(tasks):
        logger.error(
            "Invalid tasks. Ensure tasks=[{'task': .., 'args': [..], "
            "'kwargs': {..}}, ...], where 'task' is callable."
        )
        return []
    area = WorkingArea(os.path.abspath(tmpdir))
    submitter = SGETaskSubmitter(" ".join(['-N {}'.format(label), options]))
    monitor = JobMonitor(submitter)

    results = []
    area.create_areas(tasks, quiet=quiet, dill_kw=dill_kw)
    try:
        submitter.submit_tasks(area.task_paths, dryrun=dryrun, quiet=quiet)
        if not dryrun:
            results = monitor.monitor_jobs(
                sleep=sleep, request_user_input=request_resubmission_options,
            )
    except KeyboardInterrupt as e:
        submitter.killall()

    if return_files:
        return results

    results_not_files = []
    for path in results:
        with lz4.frame.open(path, 'rb') as f:
            results_not_files.append(dill.load(f))
    return results_not_files

def sge_submit_yield(
    tasks, label, tmpdir, options="-q hep.q", quiet=False, sleep=5,
    request_resubmission_options=True, dill_kw={"recurse": True},
):
    """
    Submit jobs to an SGE batch system. No monitoring is perfomed and the
    function returns an iterator over the results that yielding a list of the
    results. Jobs still running are given None in the list and the iterator
    terminates after all jobs are finished. This allows the user to perform
    some processing on partial results.

    Parameters
    ----------
    tasks : list
        A list of dictrionaries with the keys: task, args and kwargs. Each
        element is run on a node as task(*args, **kwargs).

    label : str
        Label given to the qsub submission script through -N.

    tmpdir : str
        Path to temporary directory (doesn't have to exist) where pysge stores
        job infomation. Each call will have a unique identifier in the form
        tpd_YYYYMMDD_hhmmss_xxxxxxxx. Within this directory exists all tasks in
        separate directories with a dilled file, stdout and stderr for that
        particular job.

    options : str (default = "-q hep.q")
        Additional options to pass to the qsub command. Take care since the
        following options are already in use: -wd, -V, -e, -o and -t.

    quiet : bool (default = False)
        Don't print tqdm progress bars. Other prints are controlled by logging.

    sleep : float (default = 5)
        Minimum time between queries to the batch system.

    request_resubmission_options : bool (default = True)
        When a job fails the master process will expect an stdin from the user
        to alter the submission options (e.g. to increase walltime or memory
        requested). If False it will use the original options.

    dill_kw : dict
        Kwargs to pass to dill.dump
    """

    if not _validate_tasks(tasks):
        logger.error(
            "Invalid tasks. Ensure tasks=[{'task': .., 'args': [..], "
            "'kwargs': {..}}, ...], where 'task' is callable."
        )
        return []
    area = WorkingArea(os.path.abspath(tmpdir))
    submitter = SGETaskSubmitter(" ".join(['-N {}'.format(label), options]))
    monitor = JobMonitor(submitter)

    area.create_areas(tasks, quiet=quiet, dill_kw=dill_kw)
    submitter.submit_tasks(area.task_paths, quiet=quiet)
    return monitor.request_jobs(
        sleep=sleep, request_user_input=request_resubmission_options,
    )

def sge_resume(
    label, tmpdir, options="-q hep.q", quiet=False, sleep=5,
    request_resubmission_options=True, return_files=False,
):
    """
    Resubmit jobs based on the temporary directory (with the tpd_*
    subdirectory). The original submission may terminate due to unforseen
    reasons. This will continue where everything left off.

    Parameters
    ----------
    label : str
        Label given to the qsub submission script through -N.

    tmpdir : str
        Path to temporary directory created by an sge_submit or similar.

    options : str (default = "-q hep.q")
        Additional options to pass to the qsub command. Take care since the
        following options are already in use: -wd, -V, -e, -o and -t.

    quiet : bool (default = False)
        Don't print tqdm progress bars. Other prints are controlled by logging.

    sleep : float (default = 5)
        Minimum time between queries to the batch system.

    request_resubmission_options : bool (default = True)
        When a job fails the master process will expect an stdin from the user
        to alter the submission options (e.g. to increase walltime or memory
        requested). If False it will use the original options.

    return_files : bool (default = False)
        Instead of opening the output files and loading them into python, just
        send the paths to the output files and let the user deal with them.
    """
    area = WorkingArea(os.path.abspath(tmpdir), resume=True)
    submitter = SGETaskSubmitter(" ".join(['-N {}'.format(label), options]))
    for idx in range(len(area.task_paths)):
        submitter.jobid_tasks['{}'.format(idx)] = area.task_paths[idx]
    monitor = JobMonitor(submitter)

    results = []
    try:
        results = monitor.monitor_jobs(
            sleep=sleep, request_user_input=request_resubmission_options,
        )
    except KeyboardInterrupt as e:
        submitter.killall()

    if return_files:
        return results

    results_not_files = []
    for path in results:
        with lz4.frame.open(path, 'rb') as f:
            results_not_files.append(dill.load(f))
    return results_not_files

def mp_submit(tasks, ncores=4, quiet=False):
    """
    Submit multiprocessing jobs. Each task spawns a new child process which is
    run on an available core. In multiprocessing mode the functions are
    pickled and sent to each core, hence this is good at testing small scale
    local submission before submitting to the batch. Return a list of the
    results of each job (i.e. the return values of the function calls)

    Parameters
    ----------
    tasks : list
        A list of dictrionaries with the keys: task, args and kwargs. Each
        element is run on a node as task(*args, **kwargs).

    ncores : int
        The number of cores to run on.

    quiet : bool (default = False)
        Don't print tqdm progress bars. Other prints are controlled by logging.
    """
    if not _validate_tasks(tasks):
        logger.error(
            "Invalid tasks. Ensure tasks=[{'task': .., 'args': [..], "
            "'kwargs': {..}}, ...], where 'task' is callable."
        )
        return []
    submitter = MPTaskSubmitter()
    return submitter.submit_tasks(tasks, ncores=ncores, quiet=quiet)

def local_submit(tasks, quiet=False):
    """
    Submit local jobs. Mainly for testing purposes.

    Parameters
    ----------
    tasks : list
        A list of dictrionaries with the keys: task, args and kwargs. Each
        element is run on a node as task(*args, **kwargs).

    quiet : bool (default = False)
        Don't print tqdm progress bars. Other prints are controlled by logging.
    """
    if not _validate_tasks(tasks):
        logger.error(
            "Invalid tasks. Ensure tasks=[{'task': .., 'args': [..], "
            "'kwargs': {..}}, ...], where 'task' is callable."
        )
        return []

    results = []
    pbar = tqdm(total=len(tasks), desc="Finished", disable=quiet)

    try:
        for t in tasks:
            results.append(t["task"](*t["args"], **t["kwargs"]))
            pbar.update()
    except KeyboardInterrupt:
        results = []

    pbar.close()
    return results



from .interface import (
    local_submit, mp_submit, sge_submit, sge_submit_yield, sge_resume
)

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
if not logger.handlers:
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    del handler
    del formatter
del logger



import os
import datetime
import tempfile
import lz4.frame
import glob
from tqdm.auto import tqdm
import dill
import logging
logger = logging.getLogger(__name__)

class WorkingArea(object):
    def __init__(self, path, resume=False):
        self.task_paths = None

        if resume:
            self.path = path
            self.get_areas()
        else:
            prefix = 'tpd_{:%Y%m%d_%H%M%S}_'.format(datetime.datetime.now())

            if not os.path.exists(path):
                os.makedirs(path)
            self.path = tempfile.mkdtemp(prefix=prefix, dir=os.path.abspath(path))
            if not os.path.exists(self.path):
                os.makedirs(self.path)

    def create_areas(self, tasks, quiet=False, dill_kw={"recurse": True}):
        task_paths = []
        logger.info('Creating paths in {}'.format(self.path))
        for idx, task in tqdm(
            enumerate(tasks), total=len(tasks), disable=quiet,
        ):
            package_name = 'task_{:05d}'.format(idx)
            path = os.path.join(self.path, package_name)
            if not os.path.exists(path):
                os.makedirs(path)
            file_path = os.path.join(os.path.join(path, "task.p.lz4"))
            with lz4.frame.open(file_path, 'wb') as f:
                dill.dump(task, f, **dill_kw)
            task_paths.append(path)
        self.task_paths = task_paths

    def get_areas(self):
        self.task_paths = list(glob.glob(os.path.join(self.path, "task_*")))



import setuptools

with open("README.md", 'r') as fh:
    long_description = fh.read()

with open("requirements.txt", 'r') as fh:
    requirements = fh.read().splitlines()

setuptools.setup(
    name="pysge",
    version="0.2.4",
    author="Shane Breeze",
    author_email="sdb15@ic.ac.uk",
    scripts=["pysge/pysge_worker.py", "pysge/pysge_worker.sh"],
    description="python interface to submit functions to an SGE batch cluster",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    classifiers=(
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "Development Status :: 3 - Alpha",
    ),
)
